-- --------------------------------------------------------------------
-- Title      : fixed_sin
-- Description: Test vectors for testing the fixed point package
-- This is a simple SIN function, implimented in fixed point.  Value values
-- for the input are 0 to PI/4 (0 to 90 degrees).  Accurate to "PR" bits.
-- Last Modified: $Date: 2005-07-22 10:30:35-04 $
-- RCS ID: $Id: fixed_sin.vhd,v 1.3 2005-07-22 10:30:35-04 l435385 Exp $
--
--  Created for VHDL-200X-ft, David Bishop (dbishop@vhdl.org) 
-- ---------------------------------------------------------------------------


library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.fixed_pkg.all;

entity fixed_sin is
  port (
    arg      : in  ufixed (1 downto -16);
    clk, rst : in  std_ulogic;
    res      : out ufixed (1 downto -11));
end entity fixed_sin;

architecture RTL of fixed_sin is
  constant PR : integer := -13;         -- precision
  signal arg_squ2, arg_squ3 : ufixed (3 downto PR);  -- arg_squared
  signal argu1, argu2, argu3 : ufixed (1 downto PR);  -- argument
  signal xfact2, xfact3, xfact4 : ufixed (1 downto PR);  -- X! terms
  signal term1, term2, term3 : ufixed (1 downto PR);  -- terms
  signal res1, res2, res3 : ufixed (1 downto PR);  -- results of stages
begin  -- architecture RTL

  -- purpose: performs a sin funciton
  -- sin (x) = x - (x**3/3!) + (x**5/5!) - (x**7/7!) + ....

  -- purpose: sin function stage 1
  stage1: process (clk, rst) is
  begin  -- process stage1
    if rst = '0' then                 -- asynchronous reset (active low)
      argu1 <= (others => '0');
    elsif rising_edge(clk) then         -- rising clock edge
      argu1 <= resize (arg, argu1'high, argu1'low);
    end if;
  end process stage1;

  -- purpose: sin function stage 2
  stage2: process (clk, rst) is
    constant fact3 : ufixed (0 downto PR) :=
      reciprocal(to_ufixed (2*3, -PR-1, 0));
  begin  -- process stage2
    if rst = '0' then                 -- asynchronous reset (active low)
      argu2 <= (others => '0');
      arg_squ2 <= (others => '0');      -- X**2
      xfact2 <= (others => '0');        -- X/3!
    elsif rising_edge(clk) then         -- rising clock edge
      argu2 <= argu1;
      arg_squ2 <= resize (argu1 * argu1, arg_squ2'high, arg_squ2'low);
      xfact2 <= resize (argu1 * fact3, xfact2'high, xfact2'low);
    end if;
  end process stage2;

  -- purpose: sin function stage 3
  stage3: process (clk, rst) is
    constant fact5 : ufixed (0 downto PR) :=
      reciprocal(to_ufixed (4*5, -PR-1, 0));
  begin  -- process stage2
    if rst = '0' then                 -- asynchronous reset (active low)
      argu3 <= (others => '0');
      arg_squ3 <= (others => '0');      -- X**2
      term1 <= (others => '0');         -- X**3/3!
      xfact3 <= (others => '0');        -- X**2/4*5
    elsif rising_edge(clk) then         -- rising clock edge
      argu3 <= argu2;
      arg_squ3 <= arg_squ2;
      term1 <= resize (xfact2 * arg_squ2, term1'high, term1'low);
      xfact3 <= resize (arg_squ2 * fact5, xfact3'high, xfact3'low);
    end if;
  end process stage3;

  -- purpose: sin function stage 4
  stage4: process (clk, rst) is
    constant fact7 : ufixed (0 downto PR) :=
      reciprocal(to_ufixed (6*7, -PR-1, 0));
  begin  -- process stage2
    if rst = '0' then                 -- asynchronous reset (active low)
      res1 <= (others => '0');          -- X - X**3/3!
      term2 <= (others => '0');         -- X**5/5!
      xfact4 <= (others => '0');        -- X**2/6*7
    elsif rising_edge(clk) then         -- rising clock edge
      res1 <= resize (argu3 - term1, res1'high, res1'low); -- first stage complete
      term2 <= resize (term1 * xfact3, term2'high, term2'low);
      xfact4 <= resize (arg_squ3 * fact7, xfact4'high, xfact4'low);
    end if;
  end process stage4;

  -- purpose: sin function stage 5
  stage5: process (clk, rst) is
  begin  -- process stage2
    if rst = '0' then                 -- asynchronous reset (active low)
      res2 <= (others => '0');          -- X - X**3/3! + X**5/5!
      term3 <= (others => '0');         -- X**7/7!
    elsif rising_edge(clk) then         -- rising clock edge
      res2 <= resize (res1 + term2, res2'high, res2'low); 
      term3 <= resize (term2 * xfact4, term3'high, term3'low);
    end if;   -- second stage complete
  end process stage5;

    -- purpose: sin function stage 5
  stage6: process (clk, rst) is
  begin  -- process stage2
    if rst = '0' then                 -- asynchronous reset (active low)
      res3 <= (others => '0');        -- X - X**3/3! + X**5/5! - X**7/7!
    elsif rising_edge(clk) then         -- rising clock edge
      res3 <= resize (res2 - term3, res3'high, res3'low); -- third stage complete
    end if;
  end process stage6;

      -- purpose: sin function stage 5
  outstage: process (clk, rst) is
  begin  -- process stage2
    if rst = '0' then                 -- asynchronous reset (active low)
      res <= (others => '0');
    elsif rising_edge(clk) then         -- rising clock edge
      res <= resize (res3, res'high, res'low);
    end if;
  end process outstage;

end architecture RTL;
